@@ -1,8 +1,14 @@
+import builtins
 import dis
 import sys
 import traceback
 import types
 import typing
+import math
+import json
+import os
+
+from bytecodemanipulation.CodeOptimiser import optimise_code
 
 from bytecodemanipulation.InstructionMatchers import AbstractInstructionMatcher
 from bytecodemanipulation.TransformationHelper import (
@@ -10,8 +16,9 @@ from bytecodemanipulation.TransformationHelper import (
     reconstruct_instruction,
 )
 from bytecodemanipulation.MutableCodeObject import MutableCodeObject, createInstruction
+from bytecodemanipulation.util import JUMP_ABSOLUTE
 from bytecodemanipulation.util import Opcodes
-
+from bytecodemanipulation.util import UNCONDITIONAL_JUMPS
 
 POP_JUMPS = {
     Opcodes.POP_JUMP_IF_FALSE,
@@ -788,6 +795,7 @@ class RemoveFlowBranchProcessor(AbstractBytecodeProcessor):
         while index < len(helper.instruction_listing) - 1:
             index += 1
             for index, instr in list(helper.walk())[index:]:
+                # todo: move to .json config
                 if instr.opcode in {
                     Opcodes.JUMP_IF_FALSE_OR_POP,
                     Opcodes.JUMP_IF_TRUE_OR_POP,
@@ -798,6 +806,7 @@ class RemoveFlowBranchProcessor(AbstractBytecodeProcessor):
                         match += 1
                         break
 
+                # todo: move to .json config
                 elif instr.opcode in POP_JUMPS:
                     if self.modifyAt(helper, index, match):
                         match += 1
@@ -816,13 +825,17 @@ class RemoveFlowBranchProcessor(AbstractBytecodeProcessor):
             )
 
             if self.target_jumped_branch:
-                if instr.opcode in dis.hasjabs:
+                # TODO: refactor to use .json config
+
+                if instr.opcode in JUMP_ABSOLUTE:
                     helper.insertRegion(
-                        index + 1, [createInstruction("JUMP_ABSOLUTE", instr.arg + 1)]
+                        index + 1, [createInstruction(UNCONDITIONAL_JUMPS[0], instr.arg + 1)]
                     )
+
                 else:
+                    # todo: py3.11 check for backwards jump
                     helper.insertRegion(
-                        index + 1, [createInstruction("JUMP_RELATIVE", instr.arg)]
+                        index + 1, [createInstruction(UNCONDITIONAL_JUMPS[1], instr.arg)]
                     )
 
             return True
@@ -877,3 +890,223 @@ class GlobalStaticLookupProcessor(AbstractBytecodeProcessor):
                 )
 
         helper.store()
+
+
+class SideEffectFreeMethodCallRemover(AbstractBytecodeProcessor):
+    """
+    Removes calls to function which are side effect free
+
+    Requires the methods to be statically known, by e.g. static builtins
+    """
+
+    SIDE_EFFECT_FREE_BUILTINS = {
+        min,
+        max,
+        range,
+    }
+
+    def apply(
+        self,
+        handler,
+        target: MutableCodeObject,
+        helper: BytecodePatchHelper,
+    ):
+        index = -1
+        while index < len(helper.instruction_listing) - 1:
+            index += 1
+
+            for index, instr in list(helper.walk())[index:-2]:
+                if instr.opcode == Opcodes.CALL_FUNCTION and helper.instruction_listing[index + 1].opcode == Opcodes.POP_TOP:
+                    target = next(helper.findSourceOfStackIndex(index, instr.arg))
+
+                    if target.opcode == Opcodes.LOAD_CONST:
+                        target = target.argval
+
+                        if target in self.SIDE_EFFECT_FREE_BUILTINS or ("optimiser_container" in target.__dict__ and target.optimiser_container.is_side_effect_free):
+                            helper.deleteRegion(index, index+2)
+                            helper.insertRegion(index, [createInstruction("POP_TOP")] * instr.arg)
+                            index -= instr.arg - 2
+                            break
+
+            else:
+                break
+
+        helper.store()
+
+
+class EvalAtOptimisationTime(AbstractBytecodeProcessor):
+    """
+    Helper for evaluating values at optimisation time
+
+    Require calls to be eval to be already constant functions with constant parameters
+    todo: run some normal optimisation beforehand so this can do more
+    """
+
+    OPTIMISATION_TIME_STABLE_BUILTINS = set()
+    STATIC_REF_MODULES = {}
+
+    @classmethod
+    def init_const_expr(cls):
+        with open(
+                f"{os.path.dirname(__file__)}/data/py{sys.version_info.major}.{sys.version_info.minor}_const_expressions.json") as f:
+            data = json.load(f)
+
+        # todo: make data-driven
+        cls.OPTIMISATION_TIME_STABLE_BUILTINS |= {
+            getattr(builtins, e) for e in data.setdefault("builtins", [])
+        }
+
+        for entry in data.setdefault("std_library", []):
+            module = __import__(entry["name"])
+            cls.STATIC_REF_MODULES[entry["name"]] = module
+
+            cls.OPTIMISATION_TIME_STABLE_BUILTINS |= {
+                getattr(module, e) for e in entry["items"] if hasattr(module, e)
+            }
+
+    def apply(
+        self,
+        handler,
+        target: MutableCodeObject,
+        helper: BytecodePatchHelper,
+    ):
+        optimise_code(helper)
+
+        while True:
+            for index, instr in helper.walk():
+                if instr.opcode == Opcodes.CALL_FUNCTION:
+                    target = target_opcode = next(helper.findSourceOfStackIndex(index, instr.arg))
+
+                    if target.opcode == Opcodes.LOAD_CONST:
+                        target = target.argval
+
+                        if target in self.OPTIMISATION_TIME_STABLE_BUILTINS or (hasattr(target, "__dict__") and "optimiser_container" in target.__dict__ and target.optimiser_container.is_side_effect_free):
+                            args = [next(helper.findSourceOfStackIndex(index, i)) for i in range(instr.arg - 1, -1, -1)]
+
+                            # todo: can we do more in other cases?
+                            if all(ins.opcode == Opcodes.LOAD_CONST for ins in args):
+                                value = target(*(e.argval for e in args))
+
+                                helper.instruction_listing[index] = helper.patcher.createLoadConst(value)
+
+                                for ins in args:
+                                    helper.instruction_listing[ins.offset // 2] = createInstruction("NOP")
+
+                                helper.instruction_listing[target_opcode.offset // 2] = createInstruction("NOP")
+                                break
+
+                elif instr.opcode == Opcodes.CALL_FUNCTION_KW:
+                    target = target_opcode = next(helper.findSourceOfStackIndex(index, instr.arg + 1))
+                    kw_source = next(helper.findSourceOfStackIndex(index, 0))
+
+                    if target.opcode == kw_source.opcode == Opcodes.LOAD_CONST:
+                        target = target.argval
+                        kw_args = kw_source.argval
+
+                        if target in self.OPTIMISATION_TIME_STABLE_BUILTINS or (hasattr(target, "__dict__") and "optimiser_container" in target.__dict__ and target.optimiser_container.is_side_effect_free):
+                            args = [next(helper.findSourceOfStackIndex(index, i)) for i in range(instr.arg, 0, -1)]
+
+                            # todo: can we do more in other cases?
+                            if all(ins.opcode == Opcodes.LOAD_CONST for ins in args):
+                                value = target(*(e.argval for e in args[:(len(args)-len(kw_args))]), **{key: args[len(kw_args)+i].argval for i, key in enumerate(kw_args)})
+
+                                helper.instruction_listing[index] = helper.patcher.createLoadConst(value)
+
+                                for ins in args:
+                                    helper.instruction_listing[ins.offset // 2] = createInstruction("NOP")
+
+                                helper.instruction_listing[kw_source.offset // 2] = createInstruction("NOP")
+                                helper.instruction_listing[target_opcode.offset // 2] = createInstruction("NOP")
+                                break
+
+            else:
+                break
+
+        helper.store()
+
+
+EvalAtOptimisationTime.init_const_expr()
+
+
+class StandardLibraryResolver(AbstractBytecodeProcessor):
+    def __init__(self, resolve: str):
+        self.resolve = resolve
+
+    def apply(
+        self,
+        handler,
+        target: MutableCodeObject,
+        helper: BytecodePatchHelper,
+    ):
+        library = __import__(self.resolve)
+
+        for index, instr in helper.walk():
+            if instr.opcode == Opcodes.LOAD_GLOBAL and instr.argval == self.resolve:
+                helper.instruction_listing[index] = helper.patcher.createLoadConst(library)
+
+        helper.store()
+
+
+import math, os, json, collections
+
+
+class StandardLibraryAllResolver(AbstractBytecodeProcessor):
+    STANDARD_MODULE_NAMES = EvalAtOptimisationTime.STATIC_REF_MODULES
+
+    def __init__(self, do_module_scope_lookup=True):
+        self.do_module_scope_lookup = do_module_scope_lookup
+
+    def apply(
+        self,
+        handler,
+        target: MutableCodeObject,
+        helper: BytecodePatchHelper,
+    ):
+        if self.do_module_scope_lookup:
+            imported_modules = set()
+
+            for index, instr in helper.walk():
+                if instr.opcode == Opcodes.LOAD_GLOBAL and instr.argval in target.target.__globals__:
+                    module = target.target.__globals__[instr.argval]
+                    if module.__name__ in self.STANDARD_MODULE_NAMES and self.STANDARD_MODULE_NAMES[module.__name__] == module:
+                        imported_modules.add(module.__name__)
+
+        else:
+            imported_modules = self.STANDARD_MODULE_NAMES
+
+        for index, instr in helper.walk():
+            if instr.opcode == Opcodes.LOAD_GLOBAL and instr.argval in imported_modules:
+                helper.instruction_listing[index] = helper.patcher.createLoadConst(self.STANDARD_MODULE_NAMES[instr.argval])
+
+                if helper.instruction_listing[index + 1].opcode == Opcodes.LOAD_METHOD:
+                    helper.instruction_listing[index + 1] = createInstruction("LOAD_ATTR", helper.instruction_listing[index + 1].arg)
+
+        helper.store()
+
+
+class StaticObjectAccessorResolver(AbstractBytecodeProcessor):
+    ALL_CHILDREN_OF = set(EvalAtOptimisationTime.STATIC_REF_MODULES.values())
+    SPECIFIC_CHILDREN_OF = {}
+
+    def apply(
+        self,
+        handler,
+        target: MutableCodeObject,
+        helper: BytecodePatchHelper,
+    ):
+        for index, instr in helper.walk():
+            if instr.opcode == Opcodes.LOAD_ATTR:
+                source = next(helper.findSourceOfStackIndex(index, 0))
+
+                if source.opcode == Opcodes.LOAD_CONST:
+                    if source.argval in self.ALL_CHILDREN_OF:
+                        func_usage = next(helper.findTargetOfStackIndex(index, 0))
+
+                        if func_usage.opcode == Opcodes.CALL_METHOD:
+                            helper.instruction_listing[func_usage.offset // 2] = createInstruction("CALL_FUNCTION", func_usage.arg)
+
+                        helper.instruction_listing[index] = helper.patcher.createLoadConst(getattr(source.argval, instr.argval))
+                        helper.insertRegion(index, [createInstruction("POP_TOP")])
+
+        helper.store()
+
