@@ -1,5 +1,6 @@
 import builtins
 import importlib
+import sys
 import traceback
 import types
 import typing
@@ -13,6 +14,11 @@ from bytecodemanipulation.BytecodeProcessors import (
     AbstractBytecodeProcessor,
     MethodInlineProcessor,
     Global2ConstReplace,
+    SideEffectFreeMethodCallRemover,
+    EvalAtOptimisationTime,
+    StandardLibraryResolver,
+    StandardLibraryAllResolver,
+    StaticObjectAccessorResolver,
 )
 from .BytecodeProcessors import GlobalStaticLookupProcessor
 from .InstructionMatchers import MetaArgMatcher
@@ -58,6 +64,7 @@ class _OptimiserContainer:
         self.code_walkers: typing.List[AbstractBytecodeProcessor] = []
         self.specified_locals: typing.Dict[str, typing.Type] = {}
         self.return_type: typing.Optional[typing.Type] = None
+        self.is_side_effect_free = False
 
         self.attribute_type_marks: typing.Dict[str, typing.Type] = {}
 
@@ -114,6 +121,9 @@ class _OptimiserContainer:
                 self.is_self_constant = True
 
     def optimise_target(self):
+        if "--disable-bytecode-optimisation" in sys.argv:
+            return
+
         if isinstance(self.target, types.FunctionType):
             self.optimise_method(self.target)
         else:
@@ -210,7 +220,7 @@ class _OptimiserContainer:
 def _schedule_optimisation(
     target: typing.Union[typing.Callable, typing.Type],
 ) -> _OptimiserContainer:
-    if not "optimiser_container" in target.__dict__:
+    if "optimiser_container" not in target.__dict__:
         target.optimiser_container = _OptimiserContainer(target)
         _OptimiserContainer.CONTAINERS.append(target.optimiser_container)
 
@@ -234,6 +244,34 @@ def builtins_are_static():
         optimiser.code_walkers.append(
             GlobalStaticLookupProcessor(matcher=MetaArgMatcher(_is_builtin_name))
         )
+        optimiser.code_walkers.append(
+            SideEffectFreeMethodCallRemover()
+        )
+        optimiser.code_walkers.append(
+            EvalAtOptimisationTime()
+        )
+        return target
+
+    return annotation
+
+
+def standard_library_is_safe(restriction: str = None):
+    """
+    Marker that standard library stuff is not touched by the code in an modifying way
+    """
+
+    def annotation(target: typing.Callable):
+        optimiser = _schedule_optimisation(target)
+        optimiser.code_walkers.append(
+            StandardLibraryResolver(restriction) if restriction is not None else StandardLibraryAllResolver()
+        )
+        # todo: add these only when not present, otherwise move after this
+        optimiser.code_walkers.append(
+            StaticObjectAccessorResolver()
+        )
+        optimiser.code_walkers.append(
+            EvalAtOptimisationTime()
+        )
         return target
 
     return annotation
@@ -326,7 +364,7 @@ def constant_arg(name: str):
 def forced_arg_type(name: str, type_accessor: typing.Callable, may_subclass=True):
     """
     Marks a certain arg to have that exact type, or a subclass of it when specified
-    WARNING: when passing another type, will crash as we do optimisations around that type
+    WARNING: when passing another type, may crash as we do optimisations around that type!
 
     :param name: the parameter name
     :param type_accessor: the accessor method for the type of the parameter
@@ -371,6 +409,7 @@ def constant_operation():
 
     def annotation(target: typing.Callable):
         _schedule_optimisation(target).is_constant = True
+        _schedule_optimisation(target).is_side_effect_free = True
         return target
 
     return annotation
